"""
:Author: Pierre Barbier de Reuille <pierre.barbierdereuille@gmail.com>

Module implementing non-parametric regressions using kernel smoothing methods.
"""

from __future__ import division, absolute_import, print_function
from scipy import stats
from scipy.linalg import sqrtm, solve
import scipy
import numpy as np
from .compat import irange

from .cyth import HAS_CYTHON

local_linear = None


def useCython():
    """
    Switch to using Cython methods if available
    """
    global local_linear
    if HAS_CYTHON:
        from . import cy_local_linear
        local_linear = cy_local_linear


def usePython():
    """
    Switch to using the python implementation of the methods
    """
    global local_linear
    from . import py_local_linear
    local_linear = py_local_linear

if HAS_CYTHON:
    useCython()
else:
    usePython()

from .kde import scotts_bandwidth
from .kernels import normal_kernel, normal_kernel1d


class SpatialAverage(object):
    r"""
    Perform a Nadaraya-Watson regression on the data (i.e. also called
    local-constant regression) using a gaussian kernel.

    The Nadaraya-Watson estimate is given by:

    .. math::

        f_n(x) \triangleq \frac{\sum_i K\left(\frac{x-X_i}{h}\right) Y_i}
        {\sum_i K\left(\frac{x-X_i}{h}\right)}

    Where :math:`K(x)` is the kernel and must be such that :math:`E(K(x)) = 0`
    and :math:`h` is the bandwidth of the method.

    :param ndarray xdata: Explaining variables (at most 2D array)
    :param ndarray ydata: Explained variables (should be 1D array)

    :type  cov: ndarray or callable
    :param cov: If an ndarray, it should be a 2D array giving the matrix of
        covariance of the gaussian kernel. Otherwise, it should be a function
        ``cov(xdata, ydata)`` returning the covariance matrix.
    """

    def __init__(self, xdata, ydata, cov=scotts_bandwidth):
        self.xdata = np.atleast_2d(xdata)
        self.ydata = np.atleast_1d(ydata)

        self._bw = None
        self._covariance = None
        self._inv_cov = None

        self.covariance = cov

        self.d, self.n = self.xdata.shape
        self.correction = 1.

    @property
    def bandwidth(self):
        """
        Bandwidth of the kernel. It cannot be set directly, but rather should
        be set via the covariance attribute.
        """
        if self._bw is None and self._covariance is not None:
            self._bw = np.real(sqrtm(self._covariance))
        return self._bw

    @property
    def covariance(self):
        """
        Covariance of the gaussian kernel.
        Can be set either as a fixed value or using a bandwith calculator,
        that is a function of signature ``w(xdata, ydata)`` that returns
        a 2D matrix for the covariance of the kernel.
        """
        return self._covariance

    @covariance.setter  # noqa
    def covariance(self, cov):
        if callable(cov):
            _cov = np.atleast_2d(cov(self.xdata, self.ydata))
        else:
            _cov = np.atleast_2d(cov)
        self._bw = None
        self._covariance = _cov
        self._inv_cov = scipy.linalg.inv(_cov)

    def evaluate(self, points, result=None):
        """
        Evaluate the spatial averaging on a set of points

        :param ndarray points: Points to evaluate the averaging on
        :param ndarray result: If provided, the result will be put in this
            array
        """
        points = np.atleast_2d(points).astype(self.xdata.dtype)
        #norm = self.kde(points)
        d, m = points.shape
        if result is None:
            result = np.zeros((m,), points.dtype)
        norm = np.zeros((m,), points.dtype)

        # iterate on the internal points
        for i, ci in np.broadcast(irange(self.n),
                                  irange(self._correction.shape[0])):
            diff = np.dot(self._correction[ci],
                          self.xdata[:, i, np.newaxis] - points)
            tdiff = np.dot(self._inv_cov, diff)
            energy = np.exp(-np.sum(diff * tdiff, axis=0) / 2.0)
            result += self.ydata[i] * energy
            norm += energy

        result[norm > 1e-50] /= norm[norm > 1e-50]

        return result

    def __call__(self, *args, **kwords):
        """
        This method is an alias for :py:meth:`SpatialAverage.evaluate`
        """
        return self.evaluate(*args, **kwords)

    @property
    def correction(self):
        """
        The correction coefficient allows to change the width of the kernel
        depending on the point considered. It can be either a constant (to
        correct globaly the kernel width), or a 1D array of same size as the
        input.
        """
        return self._correction

    @correction.setter  # noqa
    def correction(self, value):
        self._correction = np.atleast_1d(value)

    def set_density_correction(self):
        """
        Add a correction coefficient depending on the density of the input
        """
        kde = stats.gaussian_kde(self.xdata)
        dens = kde(self.xdata)
        dm = dens.max()
        dens[dens < 1e-50] = dm
        self._correction = dm / dens


class LocalLinearKernel1D(object):
    r"""
    Perform a local-linear regression using a gaussian kernel.

    The local constant regression is the function that minimises, for each
    position:

    .. math::

        f_n(x) \triangleq \argmin_{a_0\in\mathbb{R}}
            \sum_i K\left(\frac{x-X_i}{h}\right)
            \left(Y_i - a_0 - a_1(x-X_i)\right)^2

    Where :math:`K(x)` is the kernel and must be such that :math:`E(K(x)) = 0`
    and :math:`h` is the bandwidth of the method.

    :param ndarray xdata: Explaining variables (at most 2D array)
    :param ndarray ydata: Explained variables (should be 1D array)

    :type  cov: float or callable
    :param cov: If an float, it should be a variance of the gaussian kernel.
        Otherwise, it should be a function ``cov(xdata, ydata)`` returning the
        variance.

    """
    def __init__(self, xdata, ydata, cov=scotts_bandwidth):
        self.xdata = np.atleast_1d(xdata)
        self.ydata = np.atleast_1d(ydata)
        self.n = self.xdata.shape[0]

        self._bw = None
        self._covariance = None

        self.covariance = cov

    @property
    def bandwidth(self):
        """
        Bandwidth of the kernel.
        """
        return self._bw

    @property
    def covariance(self):
        """
        Covariance of the gaussian kernel.
        Can be set either as a fixed value or using a bandwith calculator,
        that is a function of signature ``w(xdata, ydata)`` that returns
        a single value.

        .. note::

            A ndarray with a single value will be converted to a floating
            point value.
        """
        return self._covariance

    @covariance.setter  # noqa
    def covariance(self, cov):
        if callable(cov):
            _cov = float(cov(self.xdata, self.ydata))
        else:
            _cov = float(cov)
        self._covariance = _cov
        self._bw = np.sqrt(_cov)

    def evaluate(self, points, output=None):
        """
        Evaluate the spatial averaging on a set of points

        :param ndarray points: Points to evaluate the averaging on
        :param ndarray result: If provided, the result will be put in this
            array
        """
        li2, output = local_linear.local_linear_1d(self._bw, self.xdata,
                                                   self.ydata, points, output)
        self.li2 = li2
        return output

    def __call__(self, *args, **kwords):
        """
        This method is an alias for :py:meth:`LocalLinearKernel1D.evaluate`
        """
        return self.evaluate(*args, **kwords)


class PolynomialDesignMatrix1D(object):
    def __init__(self, dim):
        self.dim = dim
        powers = np.arange(0, dim + 1).reshape((1, dim + 1))
        self.powers = powers

    def __call__(self, dX, out=None):
        return np.power(dX, self.powers, out)  # / self.frac


class LocalPolynomialKernel1D(object):
    r"""
    Perform a local-polynomial regression using a user-provided kernel
    (Gaussian by default).

    The local constant regression is the function that minimises, for each
    position:

    .. math::

        f_n(x) \triangleq \argmin_{a_0\in\mathbb{R}}
            \sum_i K\left(\frac{x-X_i}{h}\right)
            \left(Y_i - a_0 - a_1(x-X_i) - \ldots -
                a_q \frac{(x-X_i)^q}{q!}\right)^2

    Where :math:`K(x)` is the kernel such that :math:`E(K(x)) = 0`, :math:`q`
    is the order of the fitted polynomial  and :math:`h` is the bandwidth of
    the method. It is also recommended to have :math:`\int_\mathbb{R} x^2K(x)dx
    = 1`, (i.e. variance of the kernel is 1) or the effective bandwidth will be
    scaled by the square-root of this integral (i.e. the standard deviation of
    the kernel).

    :param ndarray xdata: Explaining variables (at most 2D array)
    :param ndarray ydata: Explained variables (should be 1D array)
    :param int q: Order of the polynomial to fit. **Default:** 3

    :type  cov: float or callable
    :param cov: If an float, it should be a variance of the gaussian kernel.
        Otherwise, it should be a function ``cov(xdata, ydata)`` returning
        the variance.
        **Default:** ``scotts_bandwidth``

    """
    def __init__(self, xdata, ydata, q=3, **kwords):
        self.xdata = np.atleast_1d(xdata)
        self.ydata = np.atleast_1d(ydata)
        self.n = self.xdata.shape[0]
        self.q = q

        self._kernel = None
        self._bw = None
        self._covariance = None
        self.designMatrix = None

        for n in kwords:
            setattr(self, n, kwords[n])

        if self.kernel is None:
            self.kernel = normal_kernel1d()
        if self.covariance is None:
            self.covariance = scotts_bandwidth
        if self.designMatrix is None:
            self.designMatrix = PolynomialDesignMatrix1D

    @property
    def bandwidth(self):
        """
        Bandwidth of the kernel.
        """
        return self._bw

    @bandwidth.setter  # noqa
    def bandwidth(self, bw):
        if callable(bw):
            _bw = float(bw(self.xdata, self.ydata))
        else:
            _bw = float(bw)
        self._bw = _bw
        self._covariance = _bw * _bw

    @property
    def covariance(self):
        """
        Covariance of the gaussian kernel.
        Can be set either as a fixed value or using a bandwith calculator,
        that is a function of signature ``w(xdata, ydata)`` that returns
        a single value.

        .. note::

            A ndarray with a single value will be converted to a floating
            point value.
        """
        return self._covariance

    @covariance.setter  # noqa
    def covariance(self, cov):
        if callable(cov):
            _cov = float(cov(self.xdata, self.ydata))
        else:
            _cov = float(cov)
        self._covariance = _cov
        self._bw = np.sqrt(_cov)

    @property
    def cov(self):
        """
        Covariance of the gaussian kernel.
        Can be set either as a fixed value or using a bandwith calculator,
        that is a function of signature ``w(xdata, ydata)`` that returns
        a single value.

        .. note::

            A ndarray with a single value will be converted to a floating
            point value.
        """
        return self.covariance

    @cov.setter  # noqa
    def cov(self, val):
        self.covariance = val

    @property
    def kernel(self):
        r"""
        Kernel object. Should provide the following methods:

        ``kernel.pdf(xs)``
            Density of the kernel, denoted :math:`K(x)`

        By default, the kernel is an instance of
        :py:class:`kernels.normal_kernel1d`
        """
        return self._kernel

    @kernel.setter  # noqa
    def kernel(self, val):
        self._kernel = val

    def evaluate(self, points, output=None):
        """
        Evaluate the spatial averaging on a set of points

        :param ndarray points: Points to evaluate the averaging on
        :param ndarray result: If provided, the result will be put
            in this array
        """
        xdata = self.xdata[:, np.newaxis]  # make it a column vector
        ydata = self.ydata[:, np.newaxis]  # make it a column vector
        q = self.q
        bw = self.bandwidth
        kernel = self.kernel
        designMatrix = self.designMatrix(q)
        if output is None:
            output = np.empty(points.shape, dtype=float)
        for i, p in enumerate(points):
            dX = (xdata - p)
            Wx = kernel(dX / bw)
            Xx = designMatrix(dX)
            WxXx = Wx * Xx
            XWX = np.dot(Xx.T, WxXx)
            Lx = solve(XWX, WxXx.T)[0]
            output[i] = np.dot(Lx, ydata)
        return output

    def __call__(self, *args, **kwords):
        """
        This method is an alias for :py:meth:`LocalLinearKernel1D.evaluate`
        """
        return self.evaluate(*args, **kwords)


class PolynomialDesignMatrix(object):
    """
    Class used to create a design matrix for polynomial regression
    """
    def __init__(self, dim, deg):
        self.dim = dim
        self.deg = deg

        self._designMatrixSize()

    def _designMatrixSize(self):
        """
        Compute the size of the design matrix for a n-D problem of order d.
        Can also compute the Taylors factors (i.e. the factors that would be
        applied for the taylor decomposition)

        :param int dim: Dimension of the problem
        :param int deg: Degree of the fitting polynomial
        :param bool factors: If true, the output includes the Taylor factors

        :returns: The number of columns in the design matrix and, if required,
            a ndarray with the taylor coefficients for each column of
            the design matrix.
        """
        dim = self.dim
        deg = self.deg
        init = 1
        dims = [0] * (dim + 1)
        cur = init
        prev = 0
        #if factors:
        #    fcts = [1]
        fact = 1
        for i in irange(deg):
            diff = cur - prev
            prev = cur
            old_dims = list(dims)
            fact *= (i + 1)
            for j in irange(dim):
                dp = diff - old_dims[j]
                cur += dp
                dims[j + 1] = dims[j] + dp
        #    if factors:
        #        fcts += [fact]*(cur-prev)
        self.size = cur
        #self.factors = np.array(fcts)

    def __call__(self, x, out=None):
        """
        Creates the design matrix for polynomial fitting using the points x.

        :param ndarray x: Points to create the design matrix.
            Shape must be (D,N) or (N,), where D is the dimension of
            the problem, 1 if not there.

        :param int deg: Degree of the fitting polynomial

        :param ndarray factors: Scaling factor for the columns of the design
            matrix. The shape should be (M,) or (M,1), where M is the number
            of columns of the output. This value can be obtained using
            the :py:func:`designMatrixSize` function.

        :returns: The design matrix as a (M,N) matrix.
        """
        dim, deg = self.dim, self.deg
        #factors = self.factors
        x = np.atleast_2d(x)
        dim = x.shape[0]
        if out is None:
            s = self._designMatrixSize(dim, deg)
            out = np.empty((s, x.shape[1]), dtype=x.dtype)
        dims = [0] * (dim + 1)
        out[0, :] = 1
        cur = 1
        for i in irange(deg):
            old_dims = list(dims)
            prev = cur
            for j in irange(x.shape[0]):
                dims[j] = cur
                for k in irange(old_dims[j], prev):
                    np.multiply(out[k], x[j], out[cur])
                    cur += 1
        #if factors is not None:
        #    factors = np.asarray(factors)
        #    if len(factors.shape) == 1:
        #        factors = factors[:,np.newaxis]
        #    out /= factors
        return out


class LocalPolynomialKernel(object):
    r"""
    Perform a local-polynomial regression in N-D using a user-provided kernel
    (Gaussian by default).

    The local constant regression is the function that minimises,
    for each position:

    .. math::

        f_n(x) \triangleq \argmin_{a_0\in\mathbb{R}}
            \sum_i K\left(\frac{x-X_i}{h}\right)
            \left(Y_i - a_0 - \mathcal{P}_q(X_i-x)\right)^2

    Where :math:`K(x)` is the kernel such that :math:`E(K(x)) = 0`, :math:`q`
    is the order of the fitted polynomial, :math:`\mathcal{P}_q(x)` is a
    polynomial of order :math:`d` in :math:`x` and :math:`h` is the bandwidth
    of the method.

    The polynomial :math:`\mathcal{P}_q(x)` is of the form:

    .. math::

        \mathcal{F}_d(k) = \left\{ \n \in \mathbb{N}^d \middle|
            \sum_{i=1}^d n_i = k \right\}

        \mathcal{P}_q(x_1,\ldots,x_d) = \sum_{k=1}^q
            \sum_{\n\in\mathcal{F}_d(k)} a_{k,\n}
            \prod_{i=1}^d x_i^{n_i}

    For example we have:

    .. math::

        \mathcal{P}_2(x,y) = a_{110} x + a_{101} y + a_{220} x^2 +
            a_{211} xy + a_{202} y^2

    :param ndarray xdata: Explaining variables (at most 2D array).
        The shape should be (N,D) with D the dimension of the problem
        and N the number of points. For 1D array, the shape can be (N,),
        in which case it will be converted to (N,1) array.
    :param ndarray ydata: Explained variables (should be 1D array). The shape
        must be (N,).
    :param int q: Order of the polynomial to fit. **Default:** 3
    :param callable kernel: Kernel to use for the weights. Call is
        ``kernel(points)`` and should return an array of values the same size
        as ``points``. If ``None``, the kernel will be ``normal_kernel(D)``.

    :type  cov: float or callable
    :param cov: If an float, it should be a variance of the gaussian kernel.
        Otherwise, it should be a function ``cov(xdata, ydata)`` returning
        the variance.
        **Default:** ``scotts_bandwidth``
    """
    def __init__(self, xdata, ydata, q=3, cov=scotts_bandwidth, kernel=None):
        self.xdata = np.atleast_2d(xdata)
        self.ydata = np.atleast_1d(ydata)
        self.d, self.n = self.xdata.shape
        self.q = q
        if kernel is None:
            kernel = normal_kernel(self.d)
        self.kernel = kernel

        self._bw = None
        self._covariance = None

        self.covariance = cov

    @property
    def bandwidth(self):
        """
        Bandwidth of the kernel.
        """
        return self._bw

    @property
    def covariance(self):
        """
        Covariance of the gaussian kernel.
        Can be set either as a fixed value or using a bandwith calculator,
        that is a function of signature ``w(xdata, ydata)`` that returns
        a DxD matrix.

        .. note::

            A ndarray with a single value will be converted to a floating
            point value.
        """
        return self._covariance

    @covariance.setter  # noqa
    def covariance(self, cov):
        if callable(cov):
            _cov = cov(self.xdata, self.ydata)
        else:
            _cov = np.atleast_2d(cov)
        self._covariance = _cov
        self._bw = np.real(sqrtm(_cov))

    def evaluate(self, points, output=None):
        """
        Evaluate the spatial averaging on a set of points

        :param ndarray points: Points to evaluate the averaging on
        :param ndarray output: Pre-allocated array for the result
        """
        xdata = self.xdata
        ydata = self.ydata[:, np.newaxis]  # make it a column vector
        points = np.atleast_2d(points)
        n = self.n
        q = self.q
        d = self.d
        designMatrix = PolynomialDesignMatrix(d, q)
        dm_size = designMatrix.size
        Xx = np.empty((dm_size, n), dtype=xdata.dtype)
        WxXx = np.empty(Xx.shape, dtype=xdata.dtype)
        XWX = np.empty((dm_size, dm_size), dtype=xdata.dtype)
        inv_bw = scipy.linalg.inv(self.bandwidth)
        kernel = self.kernel
        if output is None:
            output = np.empty((points.shape[1],), dtype=float)
        for i in irange(points.shape[1]):
            dX = (xdata - points[:, i:i + 1])
            Wx = kernel(np.dot(inv_bw, dX))
            designMatrix(dX, out=Xx)
            np.multiply(Wx, Xx, WxXx)
            np.dot(Xx, WxXx.T, XWX)
            Lx = solve(XWX, WxXx)[0]
            output[i] = np.dot(Lx, ydata)
        return output

    def __call__(self, *args, **kwords):
        """
        This method is an alias for :py:meth:`LocalLinearKernel1D.evaluate`
        """
        return self.evaluate(*args, **kwords)



from __future__ import division, absolute_import, print_function
import numpy as np


def local_linear_1d(bw, xdata, ydata, points, output=None):
    points = np.atleast_1d(points).astype(xdata.dtype)
    x0 = points - xdata[:, np.newaxis]
    x02 = x0 * x0
    wi = np.exp(-x02 / (2.0 * bw * bw))
    X = np.sum(wi * x0, axis=0)
    X2 = np.sum(wi * x02, axis=0)
    wy = wi * ydata[:, np.newaxis]
    Y = np.sum(wy, axis=0)
    Y2 = np.sum(wy * x0, axis=0)
    W = np.sum(wi, axis=0)
    return None, np.divide(X2 * Y - Y2 * X, W * X2 - X * X, output)



"""
Module containing utility classes and functions.
"""

from __future__ import division, absolute_import
import ctypes
import multiprocessing as mp
import numpy as np
from .compat import irange

CTYPES_CHAR_LIST = [ctypes.c_char,
                    ctypes.c_wchar
                    ]

CTYPES_INT_LIST = [ctypes.c_byte,
                   ctypes.c_short,
                   ctypes.c_int,
                   ctypes.c_long,
                   ctypes.c_longlong
                   ]

CTYPES_UINT_LIST = [ctypes.c_ubyte,
                    ctypes.c_ushort,
                    ctypes.c_uint,
                    ctypes.c_ulong,
                    ctypes.c_ulonglong
                    ]

CTYPES_FLOAT_LIST = [ctypes.c_float,
                     ctypes.c_double,
                     ctypes.c_longdouble
                     ]

CTYPES_TO_NUMPY = {ctypes.c_char: np.dtype(np.character),
                   ctypes.c_wchar: np.dtype(np.unicode),
                   }


def _get_ctype_size(ct):
    return ctypes.sizeof(ct)

for t in CTYPES_INT_LIST:
    CTYPES_TO_NUMPY[t] = np.dtype("=i{:d}".format(_get_ctype_size(t)))

for t in CTYPES_UINT_LIST:
    CTYPES_TO_NUMPY[t] = np.dtype("=u{:d}".format(_get_ctype_size(t)))

for t in CTYPES_FLOAT_LIST:
    CTYPES_TO_NUMPY[t] = np.dtype("=f{:d}".format(_get_ctype_size(t)))

NUMPY_TO_CTYPES = {CTYPES_TO_NUMPY[t]: t for t in CTYPES_TO_NUMPY}


class _dummy(object):
    pass


def _shmem_as_ndarray(raw_array, shape=None, order='C'):
    address = ctypes.addressof(raw_array)
    length = len(raw_array)
    size = ctypes.sizeof(raw_array)
    item_size = size // length

    if shape is None:
        shape = (length,)
    else:
        assert np.prod(shape) == length
    dtype = CTYPES_TO_NUMPY.get(raw_array._type_, None)
    if dtype is None:
        raise TypeError("Unknown conversion from {} to numpy type".format(raw_array._type_))
    strides = tuple(item_size * np.prod(shape[i + 1:], dtype=int) for i in irange(len(shape)))
    if order != 'C':
        strides = strides[::-1]
    d = _dummy()
    d.__array_interface__ = {'data': (address, False),
                             'typestr': dtype.str,
                             'desc': dtype.descr,
                             'shape': shape,
                             'strides': strides,
                             }
    return np.asarray(d)


def _allocate_raw_array(size, dtype):
    dtype = np.dtype(dtype)
    ct = NUMPY_TO_CTYPES.get(dtype)
    if ct is None:
        raise TypeError("Error, cannot convert numpy type {} into ctypes".format(dtype))
    return mp.RawArray(ct, int(size))


class SharedArray(object):
    def __init__(self, ra, shape=None, order='C'):
        self._ra = ra
        if ra is not None:
            self._np = _shmem_as_ndarray(self._ra, shape, order)
            self._shape = shape
        else:
            self._np = None
            self._shape = None

    def _get_np(self):
        return self._np

    def _set_np(self, content):
        self.np[:] = content

    np = property(_get_np, _set_np)

    def _get_ra(self):
        return self._ra

    ra = property(_get_ra)

    def __getinitargs__(self):
        return (self._ra, self._shape)


def array(content, dtype=None, order=None, ndmin=0):
    content = np.asarray(content)
    if dtype is None:
        dtype = content.dtype
    ra = _allocate_raw_array(np.prod(content.shape), dtype)
    shape = tuple(content.shape)
    if ndmin > len(shape):
        shape = (1,) * (ndmin - len(shape)) + shape
    sa = SharedArray(ra, shape)
    sa.np = content
    return sa


def ones(shape, dtype=float, order=None):
    ra = _allocate_raw_array(np.prod(shape), dtype)
    sa = SharedArray(ra, shape)
    sa.np = 1
    return sa


def zeros(shape, dtype=float, order=None):
    ra = _allocate_raw_array(np.prod(shape), dtype)
    sa = SharedArray(ra, shape)
    return sa



from __future__ import division, absolute_import, print_function

from .. import kernel_smoothing
import numpy as np

from ..compat import irange, izip


def fct(xs):
    """
    Function to be estimated
    """
    return np.cos(xs) + xs ** 2

methods = [kernel_smoothing.SpatialAverage,
           kernel_smoothing.LocalLinearKernel1D,
           kernel_smoothing.LocalPolynomialKernel1D,
           kernel_smoothing.LocalPolynomialKernel1D
           ]

methods_args = [{},
                {},
                {'q': 1},
                {'q': 2}
                ]


class TestConvergence1D(object):
    @classmethod
    def setupClass(cls):
        cls.xx = np.r_[0:3:256j]
        cls.yy = fct(cls.xx)
        cls.nb_samples = 100
        sizes = [2 ** i for i in range(5, 8)]
        cls.sizes = sizes
        xs = [3 * np.random.rand(cls.nb_samples * s) for s in sizes]
        noise = cls.yy.max() / 10
        ys = [fct(x) + noise * np.random.randn(len(x)) for x in xs]
        cls.xs = [x.reshape((cls.nb_samples, s)) for x, s in izip(xs, sizes)]
        cls.ys = [y.reshape((cls.nb_samples, s)) for y, s in izip(ys, sizes)]

    def convergence(self, method, args):
        # mods = [method(x, y, **args) for (x, y) in izip(self.xs, self.ys)]
        diff = np.empty(len(self.xs), dtype=float)
        xx = self.xx
        yy = self.yy
        for i, (x, y) in enumerate(izip(self.xs, self.ys)):
            res = [method(x[j], y[j], **args)(xx) for j in irange(self.nb_samples)]
            diff[i] = np.std([(yy - r) ** 2 for r in res])
        assert all(diff[1:] < diff[:-1]), 'Diff is not strictly decreasing: {}'.format(diff)

    def testConvergence(self):
        for m, a in zip(methods, methods_args):
            yield self.convergence, m, a
        kernel_smoothing.usePython()
        yield self.convergence, kernel_smoothing.LocalLinearKernel1D, {}
        kernel_smoothing.useCython()



from __future__ import division, print_function, absolute_import
import sys
import traceback
from .compat import irange, izip

nx = ny = 1
shuffled_x = shuffled_y = None
fit_args = ()
fit_kwrds = {}
fit = None
result_array = None
eval_points = None
extra_arrays = None
extra_attrs = None


def initialize_shared(nx, ny, result_array, extra_arrays, shuffled_x,
                      shuffled_y, eval_points, extra_attrs, fit,
                      fit_args, fit_kwrds):
    initialize(nx, ny, result_array.np, [ea.np for ea in extra_arrays],
               shuffled_x.np, shuffled_y.np, eval_points.np, extra_attrs,
               fit, fit_args, fit_kwrds)


def initialize(nx, ny, result_array, extra_arrays, shuffled_x, shuffled_y,
               eval_points, extra_attrs, fit, fit_args, fit_kwrds):
    globals().update(locals())


def bootstrap_result(worker, start_repeats, end_repeats):
    #print("Starting worker {} from {} to {}".format(worker, start_repeats, end_repeats))
    try:
        for i in irange(start_repeats, end_repeats):
            #print("Worker {} runs iteration {} with fit: {}".format(worker, i, fit))
            new_fit = fit(shuffled_x[..., i % nx, :], shuffled_y[i % ny, :],
                          *fit_args, **fit_kwrds)
            #print("new_fit = {}".format(new_fit))
            result_array[i + 1] = new_fit(eval_points)
            for ea, attr in izip(extra_arrays, extra_attrs):
                ea[i + 1] = getattr(new_fit, attr)
    except Exception:
        traceback.print_exc(None, sys.stderr)
        raise
    #print "Worker {} finished".format(worker)



import numpy as np
cimport numpy as np
from libc.math cimport exp

DTYPE = np.float
ctypedef np.float_t DTYPE_t

cdef void cy_li(double bw, np.ndarray[DTYPE_t, ndim=1] xdata, np.ndarray[DTYPE_t, ndim=1] ydata, np.ndarray[DTYPE_t, ndim=1] points,
                np.ndarray[DTYPE_t, ndim=1] li2, np.ndarray[DTYPE_t, ndim=1] output):
    cdef unsigned int nx = xdata.shape[0]
    cdef unsigned int npts = points.shape[0]
    cdef unsigned int i,j
    cdef np.ndarray[DTYPE_t, ndim=1] X0 = np.zeros((nx,), dtype=DTYPE)
    cdef np.ndarray[DTYPE_t, ndim=1] wi = np.zeros((nx,), dtype=DTYPE)
    cdef np.ndarray[DTYPE_t, ndim=1] bi = np.zeros((nx,), dtype=DTYPE)
    cdef double* xp  = <double*>xdata.data
    cdef double* X0p  = <double*>X0.data
    cdef double* yp  = <double*>ydata.data
    cdef double* pp  = <double*>points.data
    cdef double* wip = <double*>wi.data
    cdef double* bip = <double*>bi.data
    cdef double* Op  = <double*>output.data
    cdef double wy, x0, X, X2, lwi, lbi, sbi, li
    cdef double exp_factor = -1.0/(2.0*bw*bw)
    for j from 0 <= j < npts:
        X = 0
        X2 = 0
        for i from 0 <= i < nx:
            x0 = xp[i] - pp[j]
            X0p[i] = x0
            lwi = exp(exp_factor*x0*x0)
            wip[i] = lwi
            X += lwi*x0
            X2 += lwi*x0*x0
        sbi = 0
        for i from 0 <= i < nx:
            lbi = wip[i] * (X2 - X0p[i]*X)
            bip[i] = lbi
            sbi += lbi
        Op[j] = 0
        for i from 0 <= i < nx:
            li = bip[i] / sbi
            li2[j] += li*li
            Op[j] += li * yp[i]

def local_linear_1d(bw, xdata, ydata, points, output = None):
    bw = float(bw)
    xdata = np.ascontiguousarray(xdata, dtype=np.float)
    ydata = np.ascontiguousarray(ydata, dtype=np.float)
    points = np.ascontiguousarray(points, dtype=np.float)
    li2 = np.empty(points.shape, dtype=float)
    if output is None:
        output = np.empty(points.shape, dtype=float)
    else:
        output = np.ascontiguousarray(output, dtype=np.float)
    cy_li(bw, xdata, ydata, points, li2, output)
    return li2, output



"""
Pure Python implementation of the kernel functions
"""

from __future__ import division, absolute_import, print_function
import numpy as np
from scipy.special import erf


s2pi = np.sqrt(2.0 * np.pi)
s2 = np.sqrt(2.0)


def norm1d_pdf(z, out=None):
    """
    Full-python implementation of :py:func:`normal_kernel1d.pdf`
    """
    z = np.asfarray(z)
    if out is None:
        out = np.empty(z.shape, dtype=z.dtype)
    out = np.multiply(z, z, out)
    out *= -0.5
    np.exp(out, out)
    out /= s2pi
    return out


def norm1d_cdf(z, out=None):
    """
    Full-python implementation of :py:func:`normal_kernel1d.cdf`
    """
    z = np.asfarray(z)
    if out is None:
        out = np.empty(z.shape, dtype=z.dtype)
    np.divide(z, s2, out)
    erf(out, out)
    out *= 0.5
    out += 0.5
    return out


def norm1d_pm1(z, out=None):
    """
    Full-python implementation of :py:func:`normal_kernel1d.pm1`
    """
    z = np.asfarray(z)
    if out is None:
        out = np.empty(z.shape, dtype=z.dtype)
    np.multiply(z, z, out)
    out *= -0.5
    np.exp(out, out)
    out /= -s2pi
    return out


def norm1d_pm2(z, out=None):
    """
    Full-python implementation of :py:func:`normal_kernel1d.pm2`
    """
    z = np.asfarray(z, dtype=float)
    if out is None:
        out = np.empty(z.shape)
    np.divide(z, s2, out)
    erf(out, out)
    out /= 2
    if z.shape:
        zz = np.isfinite(z)
        sz = z[zz]
        out[zz] -= sz * np.exp(-0.5 * sz * sz) / s2pi
    elif np.isfinite(z):
        out -= z * np.exp(-0.5 * z * z) / s2pi
    out += 0.5
    return out


tricube_width = np.sqrt(35. / 243)


def tricube_pdf(z, out=None):
    z = np.asfarray(z)
    if out is None:
        out = np.empty(z.shape, dtype=z.dtype)
    out = np.multiply(z, tricube_width, out)
    sel = (out > -1) & (out < 1)
    out[~sel] = 0
    out[sel] = 70. / 81 * (1 - abs(out[sel]) ** 3.) ** 3. * tricube_width
    return out


def tricube_cdf(z, out=None):
    z = np.asfarray(z)
    if out is None:
        out = np.empty(z.shape, dtype=z.dtype)
    out = np.multiply(z, tricube_width, out)
    sel_down = out <= -1
    sel_up = out >= 1
    sel_neg = (out < 0) & (~sel_down)
    sel_pos = (out >= 0) & (~sel_up)
    out[sel_up] = 1
    out[sel_down] = 0
    out[sel_pos] = 1. / 162 * \
        (60 * (out[sel_pos] ** 7) - 7. *
         (2 * (out[sel_pos] ** 10) + 15 * (out[sel_pos] ** 4)) +
         140 * out[sel_pos] + 81)
    out[sel_neg] = 1. / 162 * \
        (60 * (out[sel_neg] ** 7) + 7. *
         (2 * (out[sel_neg] ** 10) + 15 * (out[sel_neg] ** 4)) +
         140 * out[sel_neg] + 81)
    return out


def tricube_pm1(z, out=None):
    z = np.asfarray(z)
    if out is None:
        out = np.empty(z.shape, dtype=z.dtype)
    out = np.multiply(z, tricube_width, out)
    out[out < 0] = -out[out < 0]
    sel = out < 1
    out[~sel] = 0
    out[sel] = 7 / (3564 * tricube_width) * \
        (165 * out[sel] ** 8 - 8 * (5 * out[sel] ** 11 + 33 * out[sel] ** 5) +
         220 * out[sel] ** 2 - 81)
    return out


def tricube_pm2(z, out=None):
    z = np.asfarray(z)
    if out is None:
        out = np.empty(z.shape, dtype=z.dtype)
    out = np.multiply(z, tricube_width, out)
    sel_down = out <= -1
    sel_up = out >= 1
    sel_neg = (out < 0) & ~sel_down
    sel_pos = (out >= 0) & ~sel_up
    out[sel_down] = 0
    out[sel_up] = 1
    out[sel_pos] = 35. / (tricube_width * tricube_width * 486) * \
        (4 * out[sel_pos] ** 9 - (out[sel_pos] ** 12 + 6 * out[sel_pos] ** 6) +
         4 * out[sel_pos] ** 3 + 1)
    out[sel_neg] = 35. / (tricube_width * tricube_width * 486) * \
        (4 * out[sel_neg] ** 9 + (out[sel_neg] ** 12 + 6 * out[sel_neg] ** 6) +
         4 * out[sel_neg] ** 3 + 1)
    return out


epanechnikov_width = 1. / np.sqrt(5.)


def epanechnikov_pdf(z, out=None):
    z = np.asfarray(z)
    if out is None:
        out = np.empty(z.shape, dtype=z.dtype)
    out = np.multiply(z, epanechnikov_width, out)
    sel = (out > -1) & (out < 1)
    out[~sel] = 0
    out[sel] = (.75 * epanechnikov_width) * (1 - out[sel] ** 2)
    return out


def epanechnikov_cdf(z, out=None):
    z = np.asfarray(z)
    if out is None:
        out = np.empty(z.shape, dtype=z.dtype)
    out = np.multiply(z, epanechnikov_width, out)
    sel_up = out >= 1
    sel_down = out <= -1
    out[sel_up] = 1
    out[sel_down] = 0
    sel = ~(sel_up | sel_down)
    out[sel] = .25 * (2 + 3 * out[sel] - out[sel] ** 3)
    return out


def epanechnikov_pm1(z, out=None):
    z = np.asfarray(z)
    if out is None:
        out = np.empty(z.shape, dtype=z.dtype)
    out = np.multiply(z, epanechnikov_width, out)
    sel = (out > -1) & (out < 1)
    out[~sel] = 0
    out[sel] = -3 / (16 * epanechnikov_width) * \
        (1 - 2 * out[sel] ** 2 + out[sel] ** 4)
    return out


def epanechnikov_pm2(z, out=None):
    z = np.asfarray(z)
    if out is None:
        out = np.empty(z.shape, dtype=z.dtype)
    out = np.multiply(z, epanechnikov_width, out)
    sel_up = out >= 1
    sel_down = out <= -1
    out[sel_up] = 1
    out[sel_down] = 0
    sel = ~(sel_up | sel_down)
    out[sel] = .25 * (2 + 5 * out[sel] ** 3 - 3 * out[sel] ** 5)
    return out


def normal_o4_pdf(z, out=None):
    z = np.asfarray(z)
    if out is None:
        out = np.empty(z.shape, dtype=z.dtype)
    out = norm1d_pdf(z, out)
    out *= (3 - z ** 2) / 2
    return out


def normal_o4_cdf(z, out=None):
    z = np.asfarray(z)
    if out is None:
        out = np.empty(z.shape, dtype=z.dtype)
    out = norm1d_cdf(z, out)
    out += z * norm1d_pdf(z) / 2
    return out


def normal_o4_pm1(z, out=None):
    z = np.asfarray(z)
    if out is None:
        out = np.empty(z.shape, dtype=z.dtype)
    out = norm1d_pdf(z, out)
    out -= normal_o4_pdf(z)
    return out


def normal_o4_pm2(z, out=None):
    z = np.asfarray(z)
    if out is None:
        out = np.empty(z.shape, dtype=z.dtype)
    out = np.power(z, 3, out)
    out *= norm1d_pdf(z) / 2
    return out


def epanechnikov_o4_pdf(z, out=None):
    z = np.asfarray(z)
    if out is None:
        out = np.empty(z.shape, dtype=z.dtype)
    out = np.power(z, 2., out)
    out *= -15 / 8.
    out += 9. / 8.
    out[(z < -1) | (z > 1)] = 0
    return out


def epanechnikov_o4_cdf(z, out=None):
    z = np.asfarray(z)
    if out is None:
        out = np.empty(z.shape, dtype=z.dtype)
    out = np.power(z, 3, out)
    out *= -5. / 8.
    out += (4 + 9 * z) / 8.
    out[z > 1] = 1
    out[z < -1] = 0
    return out


def epanechnikov_o4_pm1(z, out=None):
    z = np.asfarray(z)
    if out is None:
        out = np.empty(z.shape, dtype=z.dtype)
    out = np.power(z, 4, out)
    out *= -15. / 32.
    out += 1. / 32. * (18 * z ** 2 - 3)
    out[(z < -1) | (z > 1)] = 0
    return out


def epanechnikov_o4_pm2(z, out=None):
    z = np.asfarray(z)
    if out is None:
        out = np.empty(z.shape, dtype=z.dtype)
    out = np.power(z, 3, out)
    out *= .375
    out -= .375 * np.power(z, 5)
    out[(z < -1) | (z > 1)] = 0
    return out



import numpy as np
from .utils import finite
from scipy import fftpack, optimize

def variance_bandwidth(factor, xdata):
    r"""
    Returns the covariance matrix:

    .. math::

        \mathcal{C} = \tau^2 cov(X)

    where :math:`\tau` is a correcting factor that depends on the method.
    """
    data_covariance = np.atleast_2d(np.cov(xdata, rowvar=1, bias=False))
    sq_bandwidth = data_covariance * factor * factor
    return sq_bandwidth


def silverman_covariance(xdata, ydata=None, model=None):
    r"""
    The Silverman bandwidth is defined as a variance bandwidth with factor:

    .. math::

        \tau = \left( n \frac{d+2}{4} \right)^\frac{-1}{d+4}
    """
    xdata = np.atleast_2d(xdata)
    d, n = xdata.shape
    return variance_bandwidth(np.power(n * (d + 2.) / 4.,
                              -1. / (d + 4.)), xdata)


def scotts_covariance(xdata, ydata=None, model=None):
    r"""
    The Scotts bandwidth is defined as a variance bandwidth with factor:

    .. math::

        \tau = n^\frac{-1}{d+4}
    """
    xdata = np.atleast_2d(xdata)
    d, n = xdata.shape
    return variance_bandwidth(np.power(n, -1. / (d + 4.)), xdata)


def _botev_fixed_point(t, M, I, a2):
    l = 7
    I = large_float(I)
    M = large_float(M)
    a2 = large_float(a2)
    f = 2 * np.pi ** (2 * l) * np.sum(I ** l * a2 *
                                      np.exp(-I * np.pi ** 2 * t))
    for s in irange(l, 1, -1):
        K0 = np.prod(np.arange(1, 2 * s, 2)) / np.sqrt(2 * np.pi)
        const = (1 + (1 / 2) ** (s + 1 / 2)) / 3
        time = (2 * const * K0 / M / f) ** (2 / (3 + 2 * s))
        f = 2 * np.pi ** (2 * s) * \
            np.sum(I ** s * a2 * np.exp(-I * np.pi ** 2 * time))
    return t - (2 * M * np.sqrt(np.pi) * f) ** (-2 / 5)



class botev_bandwidth(object):
    """
    Implementation of the KDE bandwidth selection method outline in:

    Z. I. Botev, J. F. Grotowski, and D. P. Kroese. Kernel density
    estimation via diffusion. The Annals of Statistics, 38(5):2916-2957, 2010.

    Based on the implementation of Daniel B. Smith, PhD.

    The object is a callable returning the bandwidth for a 1D kernel.
    """
    def __init__(self, N=None, **kword):
        if 'lower' in kword or 'upper' in kword:
            print("Warning, using 'lower' and 'upper' for botev bandwidth is "
                  "deprecated. Argument is ignored")
        self.N = N

    def __call__(self, data, model, ydata=None):
        """
        Returns the optimal bandwidth based on the data
        """
        N = 2 ** 10 if self.N is None else int(2 ** np.ceil(np.log2(self.N)))
        lower = getattr(model, 'lower', None)
        upper = getattr(model, 'upper', None)
        if not finite(lower) or not finite(upper):
            minimum = np.min(data)
            maximum = np.max(data)
            span = maximum - minimum
            lower = minimum - span / 10 if not finite(lower) else lower
            upper = maximum + span / 10 if not finite(upper) else upper
        # Range of the data
        span = upper - lower

        # Histogram of the data to get a crude approximation of the density
        M = len(data)
        DataHist, bins = np.histogram(data, bins=N, range=(lower, upper), weights=model.weights)
        DataHist = DataHist / M
        DCTData = fftpack.dct(DataHist, norm=None)

        I = np.arange(1, N, dtype=int) ** 2
        SqDCTData = (DCTData[1:] / 2) ** 2
        guess = 0.1

        try:
            t_star = optimize.brentq(_botev_fixed_point, 0, guess,
                                     args=(M, I, SqDCTData))
        except ValueError:
            t_star = .28 * N ** (-.4)

        return np.sqrt(t_star) * span





"""
:Author: Pierre Barbier de Reuille <pierre.barbierdereuille@gmail.com>

Module contained a variety of small useful functions.
"""

from __future__ import division, print_function, absolute_import
from collections import OrderedDict
from keyword import iskeyword as _iskeyword
from operator import itemgetter as _itemgetter
import sys
from .compat import text_type
import numpy as np

# Find the largest float available for this numpy
if hasattr(np, 'float128'):
    large_float = np.float128
elif hasattr(np, 'float96'):
    large_float = np.float96
else:
    large_float = np.float64

def finite(val):
    return val is not None and np.isfinite(val)

def namedtuple(typename, field_names, verbose=False, rename=False):
    """Returns a new subclass of tuple with named fields.

    >>> Point = namedtuple('Point', 'x y')
    >>> Point.__doc__                   # docstring for the new class
    'Point(x, y)'
    >>> p = Point(11, y=22)             # instantiate with positional args or keywords
    >>> p[0] + p[1]                     # indexable like a plain tuple
    33
    >>> x, y = p                        # unpack like a regular tuple
    >>> x, y
    (11, 22)
    >>> p.x + p.y                       # fields also accessable by name
    33
    >>> d = p._asdict()                 # convert to a dictionary
    >>> d['x']
    11
    >>> Point(**d)                      # convert from a dictionary
    Point(x=11, y=22)
    >>> p._replace(x=100)               # _replace() is like str.replace() but targets named fields
    Point(x=100, y=22)

    """

    # Parse and validate the field names.  Validation serves two purposes,
    # generating informative error messages and preventing template injection attacks.
    if isinstance(field_names, text_type):
        # names separated by whitespace and/or commas
        field_names = field_names.replace(',', ' ').split()
    field_names = tuple(map(str, field_names))
    forbidden_fields = {'__init__', '__slots__', '__new__', '__repr__', '__getnewargs__'}
    if rename:
        names = list(field_names)
        seen = set()
        for i, name in enumerate(names):
            need_suffix = (not all(c.isalnum() or c == '_' for c in name) or _iskeyword(name)
                           or not name or name[0].isdigit() or name.startswith('_')
                           or name in seen)
            if need_suffix:
                names[i] = '_%d' % i
            seen.add(name)
        field_names = tuple(names)
    for name in (typename,) + field_names:
        if not all(c.isalnum() or c == '_' for c in name):
            raise ValueError('Type names and field names can only contain alphanumeric characters '
                             'and underscores: %r' % name)
        if _iskeyword(name):
            raise ValueError('Type names and field names cannot be a keyword: %r' % name)
        if name[0].isdigit():
            raise ValueError('Type names and field names cannot start with a number: %r' % name)
    seen_names = set()
    for name in field_names:
        if name.startswith('__'):
            if name in forbidden_fields:
                raise ValueError('Field names cannot be on of %s' % ', '.join(forbidden_fields))
        elif name.startswith('_') and not rename:
            raise ValueError('Field names cannot start with an underscore: %r' % name)
        if name in seen_names:
            raise ValueError('Encountered duplicate field name: %r' % name)
        seen_names.add(name)

    # Create and fill-in the class template
    numfields = len(field_names)
    argtxt = repr(field_names).replace("'", "")[1:-1]   # tuple repr without parens or quotes
    reprtxt = ', '.join('%s=%%r' % name for name in field_names)
    template = '''class %(typename)s(tuple):
        '%(typename)s(%(argtxt)s)' \n
        __slots__ = () \n
        _fields = %(field_names)r \n
        def __new__(_cls, %(argtxt)s):
            'Create new instance of %(typename)s(%(argtxt)s)'
            return _tuple.__new__(_cls, (%(argtxt)s)) \n
        @classmethod
        def _make(cls, iterable, new=tuple.__new__, len=len):
            'Make a new %(typename)s object from a sequence or iterable'
            result = new(cls, iterable)
            if len(result) != %(numfields)d:
                raise TypeError('Expected %(numfields)d arguments, got %%d' %% len(result))
            return result \n
        def __repr__(self):
            'Return a nicely formatted representation string'
            return '%(typename)s(%(reprtxt)s)' %% self \n
        def _asdict(self):
            'Return a new OrderedDict which maps field names to their values'
            return OrderedDict(zip(self._fields, self)) \n
        def _replace(_self, **kwds):
            'Return a new %(typename)s object replacing specified fields with new values'
            result = _self._make(map(kwds.pop, %(field_names)r, _self))
            if kwds:
                raise ValueError('Got unexpected field names: %%r' %% kwds.keys())
            return result \n
        def __getnewargs__(self):
            'Return self as a plain tuple.  Used by copy and pickle.'
            return tuple(self) \n\n''' % dict(numfields=numfields, field_names=field_names,
                                              typename=typename, argtxt=argtxt, reprtxt=reprtxt)
    for i, name in enumerate(field_names):
        template += "        %s = _property(_itemgetter(%d), " \
                    "doc='Alias for field number %d')\n" % (name, i, i)
    if verbose:
        print(template)

    # Execute the template string in a temporary namespace and
    # support tracing utilities by setting a value for frame.f_globals['__name__']
    namespace = dict(_itemgetter=_itemgetter, __name__='namedtuple_%s' % typename,
                     OrderedDict=OrderedDict, _property=property, _tuple=tuple)
    try:
        exec(template, namespace)
    except SyntaxError as e:
        raise SyntaxError(e.message + ':\n' + template)
    result = namespace[typename]

    # For pickling to work, the __module__ variable needs to be set to the frame
    # where the named tuple is created.  Bypass this step in enviroments where
    # sys._getframe is not defined (Jython for example) or sys._getframe is not
    # defined for arguments greater than 0 (IronPython).
    try:
        result.__module__ = sys._getframe(1).f_globals.get('__name__', '__main__')
    except (AttributeError, ValueError):
        pass

    return result


#
from scipy import sqrt
from numpy import finfo, asarray, asfarray, zeros

_epsilon = sqrt(finfo(float).eps)


def approx_jacobian(x, func, epsilon, *args):
    """
    Approximate the Jacobian matrix of callable function func

    :param ndarray x: The state vector at which the Jacobian matrix is desired
    :param callable func: A vector-valued function of the form f(x,*args)
    :param ndarray epsilon: The peturbation used to determine the partial derivatives
    :param tuple args: Additional arguments passed to func

    :returns: An array of dimensions (lenf, lenx) where lenf is the length
         of the outputs of func, and lenx is the number of

    .. note::

         The approximation is done using forward differences

    """
    x0 = asarray(x)
    x0 = asfarray(x0, dtype=x0.dtype)
    epsilon = x0.dtype.type(epsilon)
    f0 = func(*((x0,) + args))
    jac = zeros([len(x0), len(f0)], dtype=x0.dtype)
    dx = zeros(len(x0), dtype=x0.dtype)
    for i in range(len(x0)):
        dx[i] = epsilon
        jac[i] = (func(*((x0 + dx,) + args)) - f0) / epsilon
        dx[i] = 0.0
    return jac.transpose()



from __future__ import division, absolute_import, print_function

from .. import bootstrap

import numpy as np


class TestBootstrapMethods(object):

    @classmethod
    def setUpClass(cls):
        cls.xdata = np.r_[0:1000:129j]
        cls.res = np.r_[-4:5:129j]
        cls.centered_res = cls.res - np.mean(cls.res)
        cls.ydata = cls.xdata + cls.res

    def base_residuals(self, rep, fct, real_res, args, kwords):
        def noop(x):
            return x
        new_x, new_y = fct(noop, self.xdata, self.ydata, repeats=rep, *args, **kwords)
        used_res = new_y - new_x
        assert used_res.shape == self.xdata.shape[:-1] + (rep,) + self.xdata.shape[-1:]
        diff = (np.abs(np.subtract.outer(used_res, real_res)) < 1e-10).sum(axis=2)
        np.testing.assert_array_equal(diff, np.ones(dtype=diff.dtype, shape=diff.shape))

    def iterate_base_residuals(self, fct, real_res, *args, **kwords):
        for rep in (20, 129, 500):
            yield self.base_residuals, rep, fct, real_res, args, kwords

    def test_residuals_simple(self):
        for args in self.iterate_base_residuals(bootstrap.bootstrap_residuals, self.centered_res):
            yield args

    def test_residuals_res_fct(self):
        def residuals(y1, y0):
            return y1 - y0 + 1

        def add_residual(y, r):
            return y + r - 1

        for args in self.iterate_base_residuals(bootstrap.bootstrap_residuals,
                                                self.centered_res - 1, residuals=residuals,
                                                add_residual=add_residual):
            yield args

    def test_residuals_res_values(self):
        residuals = self.res + 1

        def add_residual(y, r):
            return y + r - 1

        for args in self.iterate_base_residuals(bootstrap.bootstrap_residuals,
                                                self.centered_res - 1, residuals=residuals,
                                                add_residual=add_residual):
            yield args

    def base_regression(self, rep, fct, args, kwords):
        def noop(x):
            return x
        new_x, new_y = fct(noop, self.xdata, self.ydata, repeats=rep, *args, **kwords)
        used_res = new_y - new_x
        assert used_res.shape == self.xdata.shape[:-1] + (rep,) + self.xdata.shape[-1:]
        diff = (np.abs(np.subtract.outer(new_y, self.ydata)) < 1e-10).sum(axis=2)
        np.testing.assert_array_equal(diff, np.ones(dtype=diff.dtype, shape=diff.shape))

    def test_regression_simple(self):
        for rep in (20, 129, 500):
            yield self.base_regression, rep, bootstrap.bootstrap_regression, (), {}


class VoidFittingFct(object):
    def __init__(self, y, *args, **kwords):
        self.args = np.array(args)
        #print(y[0], file=sys.stderr)
        self.y = y[0]

    def __call__(self, x):
        return x + self.y


class VoidFitting(object):
    def __init__(self, *args, **kwords):
        self.args = args
        self.kwords = kwords

    def __call__(self, xdata, ydata, *args, **kwords):
        assert args == self.args
        assert kwords == self.kwords
        return VoidFittingFct(ydata, *args, **kwords)


class CountShuffleSparse(object):
    def __init__(self, *args, **kwords):
        self.args = args
        self.kwords = kwords

    def __call__(self, fct, xdata, ydata, *args, **kwords):
        repeats = kwords.pop('repeats', 3000)
        assert args == self.args
        assert kwords == self.kwords
        return (xdata[..., np.newaxis, :],
                np.zeros(ydata.shape) + np.r_[1:repeats:repeats * 1j][:, np.newaxis])


class CountShuffleFull(object):
    def __init__(self, *args, **kwords):
        self.args = args
        self.kwords = kwords

    def __call__(self, fct, xdata, ydata, *args, **kwords):
        repeats = kwords.pop('repeats', 3000)
        assert args == self.args
        assert kwords == self.kwords
        return (np.repeat(xdata[..., np.newaxis, :], repeats, axis=-2),
                np.zeros(ydata.shape) + np.r_[1:repeats:repeats * 1j][:, np.newaxis])


class TestBoostrap(object):
    @classmethod
    def setUpClass(cls):
        cls.xdata = np.r_[0:1000:129j]
        cls.ydata = np.zeros((129,), dtype=float)
        cls.eval_points = np.r_[10:900:65j]

    def simple(self, rep, fit, shuffle, CI, extra_attrs, other_tests, args, kwords):
        result = bootstrap.bootstrap(fit, self.xdata, self.ydata, CI=CI, extra_attrs=extra_attrs,
                                     eval_points=self.eval_points, repeats=rep,
                                     shuffle_method=shuffle, *args, **kwords)
        np.testing.assert_array_equal(result.y_est, self.xdata)
        np.testing.assert_array_equal(result.y_eval, self.eval_points)
        assert len(result.CIs) == 1 + len(extra_attrs)
        for ci in result.CIs:
            assert ci.shape[:-1] == (len(CI), 2)
        assert result.CIs[0].shape[-1] == len(self.eval_points)

        if result.shuffled_xs is not None:
            assert result.shuffled_xs.shape[-2] in (1, rep)
            assert result.shuffled_ys.shape[-2] in (1, rep)
            assert result.full_results.shape == (rep + 1, len(self.eval_points))
            expected = self.eval_points + np.r_[0:rep:(rep + 1) * 1j][:, np.newaxis]
            np.testing.assert_array_equal(result.full_results,
                                          expected)
        other_tests(result)

    def iter_simple(self, fit, shuffle, CI, extra_attrs=(),
                    other_tests = lambda x: None, *args, **kwords):
        for rep in (20, 129, 500):
            for worker in (1, None, 2, 4):
                kw = dict(kwords)
                kw['nb_workers'] = worker
                yield self.simple, rep, fit, shuffle, CI, extra_attrs, other_tests, args, kw

    def test_simple_sparse(self):
        for arg in self.iter_simple(VoidFitting(1, 2, a=3, b=4),
                                    CountShuffleSparse(10, 12, d=4, e=6), CI=(95,),
                                    shuffle_args = (10, 12), shuffle_kwrds = {'d': 4, 'e': 6},
                                    fit_args=(1, 2), fit_kwrds = {'a': 3, 'b': 4}):
            yield arg

    def test_simple_attrs(self):
        for arg in self.iter_simple(VoidFitting(1, 2, a=3, b=4),
                                    CountShuffleSparse(10, 12, d=4, e=6),
                                    CI=(95,), extra_attrs = ('args',),
                                    shuffle_args = (10, 12),
                                    shuffle_kwrds = {'d': 4, 'e': 6},
                                    fit_args=(1, 2), fit_kwrds = {'a': 3, 'b': 4}):
            yield arg

    def test_simple_full(self):
        for arg in self.iter_simple(VoidFitting(), CountShuffleSparse(),
                                    CI=(95,), full_results=True):
            yield arg



# -*- coding: utf-8 -*-
"""
Created on Tue Jul 31 20:11:03 2012

@author: Pierre Barbier de Reuille <pierre.barbierdereuille@gmail.com>
"""

from __future__ import absolute_import, print_function
import os
import numpy
try:
    import pyximport
    HAS_CYTHON = True
except ImportError:
    HAS_CYTHON = False

if HAS_CYTHON:
    USE_MINGW=False
    if os.name == 'nt':
        if 'CPATH' in os.environ:
            os.environ['CPATH'] = os.environ['CPATH'] + numpy.get_include()
        else:
            os.environ['CPATH'] = numpy.get_include()

        if USE_MINGW:
            if 'PATH' in os.environ:
                os.environ['PATH'] = os.environ['PATH'] + r';C:\MinGW\bin'
            else:
                os.environ['PATH'] = r'C:\MinGW\bin'
            mingw_setup_args = {'options': {'build_ext': {'compiler': 'mingw32'}}}

        pyximport.install(setup_args=mingw_setup_args, reload_support=True)

    elif os.name == 'posix':
        extra_flags = '-I' + numpy.get_include()
        os.environ['CFLAGS'] = " ".join([os.environ.get('CFLAGS', ''),
                                         extra_flags])
        os.environ['CXXFLAGS'] = " ".join([os.environ.get('CXXFLAGS', ''),
                                           extra_flags])

        pyximport.install(reload_support=True)

